#!/usr/bin/env python
# coding: utf-8

AUTHOR = "Christian Kokoska"

from flask import g

class InvalidFieldName(Exception):
    pass

class NoParameters(Exception):
    pass

def get_order(filter):
    """
    Helper function which provides the order query
    """
    order = ""
    for key, value in filter.items():
        if key == "order_by":
            if value.startswith("-"):
                value = value[1:]
                order = "ORDER BY %s " % (value) + "DESC"
            else:
                order = "ORDER BY %s " % (value) + "ASC"
    return order


def get_objects(cls, filter):
    """
    Helper function to return the fetched values in the right order
    """
    objects = []
    mapped = {}
    for item in cls.adaptor.get_by_filter(filter, cls.tablename, cls.fields.keys()):
        object = cls(*item)
        mapped = zip(cls.fields.keys(), item)
        for item in mapped:
            setattr(object, item[0], item[1])
        objects.append(object)
    return objects


class DbObject(object):
    """
    The class from which the classes in the application inherits
    """
    def __init__(self):
        self.adaptor = None

    @classmethod
    def all(cls, **kwargs):
        filter = "1=1 " + get_order(kwargs)
        return get_objects(cls, filter)

    @classmethod
    def get(cls, *args, **kwargs):
        if not kwargs:
            raise NoParameters('No parameters provided')
        return cls.filter(*args, **kwargs)[0]

    @classmethod
    def filter(cls, *args, **kwargs):
        if not kwargs:
            raise NoParameters('No parameters provided')
        filters = []
        order = ""

        for key, value in kwargs.iteritems():
            if key.split('__')[0] == "max":
                key_value = cls.adaptor.get_key_value(key.split('__')[1], value, cls.tablename, cls.fields)

                filters.append("%s<=%s" % key_value)
            elif key.split('__')[0] == "min":
                key_value = cls.adaptor.get_key_value(key.split('__')[1], value, cls.tablename, cls.fields)

                filters.append("%s>=%s" % key_value)
            elif key == "order_by":
                continue
            else:
                key_value = cls.adaptor.get_key_value(key, value, cls.tablename, cls.fields)
                filters.append("%s=%s" % key_value)
        filter = (" AND ").join(filters)
        if not filter:
            filter = "1=1"
        filter += " " + get_order(kwargs)
        return get_objects(cls, filter)

    @classmethod
    def create(cls, *args, **kwargs):
        """
        This method can be called to create an object

        :param class: The class which calls this method
        :return: The freshly created object
        """
        return cls.adaptor.create(cls, cls.tablename, cls.fields.keys(), kwargs)

    def save(self, *args, **kwargs):
        """
        Function can be called to create or update an object
        """
        if self.id:
            self.update(self.tablename, self.fields.keys(), kwargs)
        else:
            for fieldname in self.fields.keys():
                kwargs[fieldname] = getattr(self, fieldname)
            self.create(self, self.tablename, self.fields.keys(), **kwargs)

    def update(self, *args, **kwargs):
        return self.adaptor.update(self.tablename, self.fields.keys(), self)

    def delete(self, *args, **kwargs):
        return self.adaptor.delete(self.tablename, self)

    @classmethod
    def pure(cls, *args):
        """
        Function which can be used to select items from the database by pure SQL

        :param query: A pure SQL select query
        :return: An object of the calling class
        """
        return [cls(*item) for item in cls.adaptor.get_by_sql(args[0])]

    @classmethod
    def raw(cls, query, *args):
        """
        Executes a query
        """
        con = cls.adaptor.get_connection()
        cur = con.cursor()
        cur.execute(query)
        try:
            fetch = cur.fetchall()
        except:
            fetch = None
        con.commit()
        return fetch

    @classmethod
    def create_table(cls):
        """
        Helper function to create tables based on their fieldnames and tablename
        """
        return cls.adaptor.create_table(cls)


class Adaptor():
    """
    Class to provide some common methods which can be used by more than one
    child adaptor
    """

    def get_key_value(self, key, value, *args):
        tablename = args[0]
        fields = args[1]
        if fields[key] == 'char':
            value = "'%s'" % value
        return (key, value)

    def update(self, tablename, fieldnames, *args):
        _object = args[0]
        settings = ""
        for fieldname in fieldnames:
            settings += ", " + fieldname + "=%s"
        settings = settings[2:]

        query = "UPDATE %s SET %s WHERE id = %%s" % (tablename, settings)

        values = []
        for fieldname in fieldnames:
            values.append(getattr(_object, fieldname))
        values.append(_object.id)

        con = self.get_connection()
        con.cursor().execute(query, values)
        con.commit()

    def delete(self, tablename, *args):
        _object = args[0]
        query = "DELETE FROM %s WHERE id=%%s" % (tablename)
        con = self.get_connection()
        con.cursor().execute(query, [_object.id])
        con.commit()


class PostgreSQLAdaptor(Adaptor):

    def __init__(self, host, database, username, password):
        self.host = host
        self.database = database
        self.username = username
        self.password = password

    def get_connection(self):
        import psycopg2
        psycopg2.extensions.register_type(psycopg2.extensions.UNICODE)
        con = psycopg2.connect(host=self.host,
                              database=self.database,
                              user=self.username,
                              password=self.password)
        return con

    def get_by_sql(self, query, *args):
        cur = self.get_connection().cursor()
        cur.execute(query)
        return cur.fetchall()

    def get_by_filter(self, filter, *args):
        tablename = args[0]
        fieldnames = args[1]
        query = "SELECT %s FROM %s WHERE %s" % (', '.join(fieldnames),
                                                tablename,
                                                filter)
        return self.get_by_sql(query, *args)

    def create(self, cls, tablename, fieldnames, *args, **kwargs):
        empty_fields = set(fieldnames)
        field_values = {}

        for key, value in args[0].iteritems():
            if key not in fieldnames:
                raise InvalidFieldName
            else:
                empty_fields -= set([key])
                field_values[key] = value

        for key in empty_fields:
            field_values[key] = None

        field_values.pop('id')
        fieldnames.remove('id')

        query = "INSERT INTO %s (%s) VALUES (%s) RETURNING id" % (
                                       tablename,
                                       ', '.join(field_values.keys()),
                                       ', '.join(['%s'] * len(fieldnames)))

        fieldnames.insert(0, 'id')

        con = self.get_connection()
        cursor = con.cursor()
        cursor.execute(query, field_values.values())
        _id = cursor.fetchone()
        con.commit()
        return cls.get(id=_id[0])

    def create_table(self, cls):
        fields = ""
        for key, value in cls.fields.items():
            if key == 'id':
                fields += "id SERIAL PRIMARY KEY, "
            if value == 'char':
                fields += "%s VARCHAR(200) DEFAULT NULL, " % key
            if value == 'int' and not key == 'id':
                fields += "%s INT DEFAULT 0, " % key
            if value == 'text':
                fields += "%s TEXT DEFAULT NULL, " % key
            if value == 'bool':
                fields += "%s BOOLEAN, " % key
            if value == 'date':
                fields += "%s TIMESTAMP DEFAULT NULL, " % key
        query = """CREATE TABLE IF NOT EXISTS %s (%s)""" % (cls.tablename,
                                                              fields[:-2])
        cls.raw(query)
        return True


class MySQLAdaptor(Adaptor):

    def __init__(self, host, database, username, password):
        self.host = host
        self.database = database
        self.username = username
        self.password = password

    def get_connection(self):
        import MySQLdb
        con = MySQLdb.connect(host=self.host,
                              db=self.database,
                              user=self.username,
                              passwd=self.password,
                              use_unicode=True,
                              charset='utf8')
        return con

    def get_by_sql(self, query, *args):
        cur = self.get_connection().cursor()
        cur.execute(query)
        return cur.fetchall()

    def get_by_filter(self, filter, *args):
        tablename = args[0]
        fieldnames = args[1]
        query = "SELECT %s FROM %s WHERE %s" % (', '.join(fieldnames),
                                                tablename,
                                                filter)
        return self.get_by_sql(query, *args)

    def get_last_inserted_id(self, con):
        cur = con.cursor()
        cur.execute("""SELECT last_insert_id()""")
        return cur.fetchone()[0]

    def create(self, cls, tablename, fieldnames, *args, **kwargs):
        empty_fields = set(fieldnames)
        field_values = {}

        for key, value in args[0].iteritems():
            if key not in fieldnames:
                raise InvalidFieldName
            else:
                empty_fields -= set([key])
                field_values[key] = value

        for key in empty_fields:
            field_values[key] = None

        query = "INSERT INTO %s (%s) VALUES (%s)" % (
                                            tablename,
                                            ', '.join(field_values.keys()),
                                            ', '.join(['%s'] * len(fieldnames)))

        con = self.get_connection()
        con.cursor().execute(query, field_values.values())
        con.commit()
        return cls.get(id=self.get_last_inserted_id(con))

    def create_table(self, cls):
        fields = ""
        pk = ""
        for key, value in cls.fields.items():
            if key == 'id':
                fields += "id INT NOT NULL AUTO_INCREMENT, "
                pk = ", PRIMARY KEY (id)"
            if value == 'char':
                fields += "%s VARCHAR(200) DEFAULT NULL, " % key
            if value == 'int' and not key == 'id':
                fields += "%s INT DEFAULT 0, " % key
            if value == 'text':
                fields += "%s TEXT DEFAULT NULL, " % key
            if value == 'bool':
                fields += "%s BOOLEAN, " % key
            if value == 'date':
                fields += "%s DATETIME DEFAULT NULL, " % key
        query = """CREATE TABLE IF NOT EXISTS %s (%s %s)""" % (cls.tablename,
                                                              fields[:-2],
                                                              pk)
        cls.raw(query)
        return True


class SQLiteAdaptor(Adaptor):

    def __init__(self, database):
        self.database = database

    def get_connection(self):
        import sqlite3
        con = sqlite3.connect(self.database)
        return con

    def get_key_value(self, key, value, *args):
        tablename = args[0]
        fields = args[1]
        if fields[key] == 'bool':
            if value:
                value = 1
            else:
                value = 0
        if fields[key] == 'char':
            value = "'%s'" % value
        return (key, value)

    def get_by_sql(self, query, *args):
        cur = self.get_connection().cursor()
        cur.execute(query)
        return cur.fetchall()

    def get_by_filter(self, filter, *args):
        tablename = args[0]
        fieldnames = args[1]
        query = "SELECT %s FROM %s WHERE %s" % (', '.join(fieldnames),
                                                tablename,
                                                filter)
        return self.get_by_sql(query, *args)

    def get_last_inserted_id(self, con):
        cur = con.cursor()
        cur.execute("""SELECT last_insert_rowid()""")
        return cur.fetchone()[0]

    def create(self, cls, tablename, fieldnames, *args, **kwargs):
        empty_fields = set(fieldnames)
        field_values = {}

        for key, value in args[0].iteritems():
            if key not in fieldnames:
                raise InvalidFieldName
            else:
                empty_fields -= set([key])
                field_values[key] = value

        for key in empty_fields:
            field_values[key] = None

        query = "INSERT INTO %s (%s) VALUES (%s)" % (
                                       tablename,
                                       ', '.join(field_values.keys()),
                                       ', '.join(['?'] * len(fieldnames)))

        values = []
        for value in field_values.values():
            values.append(value)

        con = self.get_connection()
        con.cursor().execute(query, values)
        con.commit()
        return cls.get(id=self.get_last_inserted_id(con))

    def update(self, tablename, fieldnames, *args):
        _object = args[0]

        settings = ""
        for fieldname in fieldnames:
            settings += ", " + fieldname + "=?"
        settings = settings[2:]

        values = []
        for fieldname in fieldnames:
            values.append(str(getattr(_object, fieldname)))

        query = "UPDATE %s SET %s WHERE id = %s" % (tablename, settings, _object.id)

        con = self.get_connection()
        con.cursor().execute(query, values)
        con.commit()

    def delete(self, tablename, *args):
        _object = args[0]
        query = "DELETE FROM %s WHERE id=?" % (tablename)
        con = self.get_connection()
        con.cursor().execute(query, [_object.id])
        con.commit()

    def create_table(self, cls):
        fields = ""
        for key, value in cls.fields.items():
            if key == 'id':
                fields += "id INTEGER PRIMARY KEY AUTOINCREMENT, "
            if value == 'char':
                fields += "%s VARCHAR(200) DEFAULT NULL, " % key
            if value == 'int' and not key == 'id':
                fields += "%s INTEGER DEFAULT 0, " % key
            if value == 'text':
                fields += "%s TEXT DEFAULT NULL, " % key
            if value == 'bool':
                fields += "%s BOOLEAN, " % key
            if value == 'date':
                fields += "%s TIMESTAMP DEFAULT NULL, " % key

        query = """CREATE TABLE IF NOT EXISTS %s (%s)""" % (cls.tablename, fields[:-2])
        cls.raw(query)
        return True


class MyDbAdaptor(Adaptor):
    """
    Pet project
    """

    def __init__(self, database):
        self.database = database

    def get_connection(self):
        import db_driver
        con = db_driver.connect(self.database)
        return con

    def get_key_value(cls, key, value, *args):
        tablename = args[0]
        fields = args[1]
        list = ['int', 'date', 'bool']
        if fields[key] in list:
            value = "'%s'" % value

        if fields[key] == 'char':
            value = "'%s'" % value
        return (key, value)

    def get_by_sql(self, query, *args):
        con = self.get_connection()
        return con.execute(query)

    def get_by_filter(self, filter, *args):
        tablename = args[0]
        fieldnames = args[1]
        query = "SELECT %s FROM %s WHERE %s" % (', '.join(fieldnames),
                                                tablename,
                                                filter)
        return self.get_by_sql(query, *args)

    def update(self, tablename, fieldnames, *args):
        _object = args[0]
        settings = ""
        for fieldname in fieldnames:
            settings += ", " + fieldname + "=%s"
        settings = settings[2:]

        query = "UPDATE %s SET %s WHERE id=%%s" % (tablename, settings)

        values = []
        for fieldname in fieldnames:
            values.append(getattr(_object, fieldname))
        values.append(_object.id)

        con = self.get_connection()
        con.execute(query % tuple(values))

    def create(self, cls, tablename, fieldnames, *args, **kwargs):
        empty_fields = set(fieldnames)
        field_values = {}

        for key, value in args[0].iteritems():
            if key not in fieldnames:
                raise InvalidFieldName
            else:
                empty_fields -= set([key])
                field_values[key] = value

        for key in empty_fields:
            field_values[key] = None

        query = "INSERT INTO %s (%s) VALUES (%s)" % (
                                       tablename,
                                       ', '.join(field_values.keys()),
                                       ', '.join(['%s'] * len(fieldnames)))

        values = []
        for value in field_values.values():
            values.append(value)

        con = self.get_connection()
        return cls.get(id=con.execute(query % tuple(values)))

    def delete(self, tablename, *args):
        _object = args[0]
        query = "DELETE FROM %s WHERE id=%s" % (tablename, _object.id)
        con = self.get_connection()
        con.execute(query)



# coding: utf-8
from setuptools import setup


setupargs = {
    'name': 'myorm',
    'description': 'Provides a simple ORM for MySQL, PostgreSQL and SQLite.',

    'license': 'GPLv3',
    'version': '0.0.6',

    'packages': ['myorm'],

    'author': 'Christian Kokoska',
    'author_email': 'christian@eternalconcert.de',
}

if __name__ == '__main__':
    setup(**setupargs)



#!/usr/bin/env python
# coding: utf-8

import ast
import os


class Connection(object):

    def __init__(self, database):
        self.database = database

    def get_sfr(self, query):
        """
        Return a dict for a standard SELECT-FROM-WHERE query
        """

        def get_select_fields(query):
            """
            Extracts the select fields from the query
            """
            return [field for field in query.split('FROM')[0][7:-1].split(', ')]

        def get_select_filter(query):
            """
            Extracts the filter values from the query
            """
            try:
                filter = query.split('WHERE')[1][1:]
                if filter.endswith(";"):
                    filter = filter[:-1]
            except:
                filter = None
            return filter

        def get_select_table(query):
            """
            Extracts the table from the query
            """
            select_filter = get_select_filter(query)
            if select_filter:
                length = len(select_filter) + 7
                return query.split('FROM')[1][1:-length]
            else:
                return query.split('FROM')[1][1:]

        return {'fields': (get_select_fields(query)),
                'table': (get_select_table(query)),
                'filter' :(get_select_filter(query))}


    def get_insert_values_table_fields(self, query):

        def get_insert_table(query):
            table = query.split("(")[0][12:-1]
            return table

        def get_insert_fields_and_values(query):
            len1 = len(query.split("(")[0]) + 1
            len2 = len(query.split(")")[0])
            fields = query[len1:len2]
            values = query.split("(")[2][:-1]
            fields_and_values = dict(zip(fields.split(", "), values.split(", ")))
            return fields_and_values


        return {'table': (get_insert_table(query)),
                'fields_and_values': (get_insert_fields_and_values(query))}

    def get_delete_lines(self, query):

        def get_line_data(query):
            line_data = {}
            filter = query.split("WHERE ")[1]
            line_data['key'] = filter.split('=')[0]
            line_data['value'] = filter.split('=')[1]
            table = query.split("FROM")[1][1:]
            table = table.split('WHERE')[0]
            line_data['table'] = table
            return line_data

        db_content = self.get_db_content()
        line_data = get_line_data(query)
        table_lines = []
        for line in db_content.split('\n'):
            if line.startswith(line_data['table']):
                table_lines.append(line)

        lines = []
        for line in table_lines:
            check_line = ast.literal_eval(line.split(" = ")[1])
            for key, value in check_line.items():
                if key == line_data['key'] and value == line_data['value']:
                    lines.append(line)
        return lines

    def get_update_line_data(self, query):
        line_data = {}
        filter = query.split("WHERE ")[1]
        line_data['key'] = filter.split('=')[0]
        line_data['value'] = filter.split('=')[1]
        table = query.split(" SET")[0][7:]
        table = table.split('WHERE')[0]
        line_data['table'] = table
        settings = query.split("SET ")[1].split(" WHERE")[0]
        line_data['settings'] = {}
        for item in settings.split(", "):
            line_data['settings'][item.split("=")[0]] = item.split("=")[1]
        return line_data

    def get_update_line(self, query):
        db_content = self.get_db_content()
        line_data = self.get_update_line_data(query)
        table_lines = []
        for line in db_content.split('\n'):
            if line.startswith(line_data['table']):
                table_lines.append(line)
        lines = []
        for line in table_lines:
            check_line = ast.literal_eval(line.split(" = ")[1])
            for key, value in check_line.items():
                if key == line_data['key'] and value == line_data['value']:
                    lines.append(line)
        return lines


    def execute(self, query):
        query = query.strip()

        class InvalidQuery(Exception):
            pass

        def select_query(self, query):

            filter = self.get_sfr(query)['filter']
            try:
                int(filter.split("=")[0])
                filter = None
            except:
                pass

            if filter:
                return self.select_by_fields_and_filter(
                                                  self.get_sfr(query)['table'],
                                                  self.get_sfr(query)['fields'],
                                                  filter)
            else:
                return self.get_all(self.get_sfr(query)['table'],
                                    self.get_sfr(query)['fields'],)

        def insert_query(self, query):
            data = self.get_insert_values_table_fields(query)
            id = self.get_max_id_for_table(data['table']) + 1
            data['fields_and_values']['id'] = str(id)
            line = "%s = %s\n" % (data['table'], data['fields_and_values'])
            db = open(self.database, 'a')
            db.writelines(line)
            return id

        def clean_lines():
            content = self.get_db_content()
            clean = ""
            for line in content.split('\n'):
                if line[:1]:
                    clean += line + "\n"
            self.save_db_content(clean)

        def update_query(self, query):
            data = self.get_update_line_data(query)
            lines = self.get_update_line(query)
            content = self.get_db_content()

            for line in lines:
                settings = ast.literal_eval(line.split(" = ")[1])
                for key, value in data['settings'].items():
                    settings[key] = value
                new_line = "%s = %s\n" % (data['table'], settings)
                content = content.replace(line, new_line)
            self.save_db_content(content)
            clean_lines()

        def delete_query(self, query):
            lines  = self.get_delete_lines(query)
            for line in lines:
                content = self.get_db_content()
                content = content.replace(line, "")
                self.save_db_content(content)
            clean_lines()

        query_type = query[:6].upper()

        if query_type == "SELECT":
            return select_query(self, query)
        elif query_type == "INSERT":
            return insert_query(self, query)
        elif query_type == "UPDATE":
            return update_query(self, query)
        elif query_type == "DELETE":
            return delete_query(self, query)
        else:
            raise InvalidQuery("Invalid SQL provided")

    def get_max_id_for_table(self, tablename):
        ids = self.get_all(tablename, fields=['id'])
        if ids:
            return max([int(id[0]) for id in ids])
        else:
            return 0

    def get_db_content(self):
        db = open(self.database, 'r')
        return db.read()

    def save_db_content(self, content):
        db = open(self.database, 'w')
        db.write(content)
        db.close()

    def get_table_as_dict(self, tablename):
        content = self.get_db_content()
        dict = {}
        count = 0
        for line in content.split('\n'):
            if line.startswith(tablename):
                dict[count] = ast.literal_eval(line.split(" = ")[1])
                count += 1
        return dict

    def get_all(self, tablename, fields):
        """
        Selects all database objects from one table
        """
        all = []

        for _object in self.get_table_as_dict(tablename).values():

            element = {}
            for key, value in _object.items():
                element[key] = value

            order = {}
            order_position = 0
            for field in fields:
                order[order_position] = field
                order_position += 1

            sorted_fields = []

            for key, value in order.items():
                sorted_fields.append((value, element[value]))

            element = []

            for item in sorted_fields:
                item = item[1]
                element.append(item)

            all.append(tuple(element))
        return all

    def select_by_fields_and_filter(self, tablename, fields, filter):
        """
        Selects database objects by fields and filter
        """
        try:
            filter_key = filter.split('=')[0]
            filter_value = filter.split('=')[1][1:-1]
        except:
            key = filter.split("=")[0]
            value = filter.split('=')[1]
            try:
                int(key) == int(value)
            except:
                raise InvalidQuery("Something went terriblly wrong")

        all = []
        filtered = []
        for _object in self.get_table_as_dict(tablename).values():

            if _object[filter_key] == filter_value:
                filtered.append(_object)

        for selected in filtered:
            element = {}
            for key, value in selected.items():
                element[key] = value

            order = {}
            order_position = 0
            for field in fields:
                order[order_position] = field
                order_position += 1

            sorted_fields = []
            for key, value in order.items():
                sorted_fields.append((value, element[value]))

            element = []
            for item in sorted_fields:
                item = item[1]
                element.append(item)

            all.append(tuple(element))
        return all

def connect(database):
    return Connection(database)


def console(database):
    con = connect(database)
    while True:
        query = raw_input('MyDb#: ')
        try:
            for item in con.execute(query):
                print ", ".join(item)
        except:
            pass

if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('database')
    args = parser.parse_args()
    console(args.database)